# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from six import iteritems

class Resource(object): # pylint: disable=too-few-public-methods
    """
    Base object for shenanigans.
    """
    def __init__(self, **kwargs):
        for prop, val in iteritems(kwargs):
            setattr(self, prop, val)
        self.validate()

    def validate(self):
        """
        Validate resource with defined meta data.
        """
        meta = getattr(self, 'meta', None)
        if meta and meta.get('required_fields'):
            required_fields = meta.get('required_fields')
            for field in required_fields:
                if isinstance(field, (list, tuple)):
                    field, types = field
                    if not hasattr(self, field):
                        raise AttributeError('Resource missing required field "{}"'.format(field))
                    if not isinstance(getattr(self, field), types):
                        raise ValueError(
                            'Field "{}" is not of type "{}" as expected'.format(field, types))
                else:
                    if not hasattr(self, field):
                        raise AttributeError('Resource missing required field "{}"'.format(field))



import json
import time
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import NoSuchElementException, ElementNotVisibleException, InvalidElementStateException
from selenium.webdriver.support.wait import WebDriverWait
from six import string_types
from selenium.webdriver.remote.webelement import WebElement
import uuid
import re
from selenium.webdriver.common.action_chains import ActionChains

from tractordriver.resource import Resource


class Element(WebElement):
    def js_click(self):
        """
        :Description: Execute the `click` event on the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        """
        self._parent.execute_script('arguments[0].click()', self)
        return self

    @classmethod
    def __type2js(cls, value):
        """
        :Description: Convert python value to executable javascript value by type.
        :param value: Value to transform.
        :type value: None, bool, int, float, string
        :return: string
        """
        if value is None:
            return 'null'
        elif isinstance(value, bool):
            return 'false' if not value else 'true'
        elif isinstance(value, (int, float)):
            return '%s' % value
        elif isinstance(value, dict):
            return json.dumps(dict)
        return '"%s"' % value

    @classmethod
    def __type2python(cls, value):
        """
        :Description: Convert javascript value to python value by type.
        :param value: Value to transform.
        :type value: None, bool, int, float, string
        :return: None, bool, int, float, string
        """
        if isinstance(value, string_types):
            if value is 'null':
                return None
            elif value in ('true', 'false'):
                return False if value == 'false' else True
            elif value.replace('.', '', 1).isdigit():
                return eval(value)
            elif value.startswith('{') and value.endswidth('}'):
                try:
                    return json.loads(value)
                except ValueError:
                    return value
        return value

    def js_wait(self, condition, interval, *args):
        """
        :Description: Create an interval in vm.window, will clear interval after condition met.
        :param condition: Condition in javascript to pass to interval.
        :example: '$el.innerText == "cheesecake"'
        :example: '$el[0].disabled && $el[1].disabled'
        :type condition: string
        :param interval: Time in milliseconds to execute interval.
        :type interval: int or float
        :param *args: WebElement or selector of condition element.
        :type *args: tuple
        :return: string
        """
        hid = lambda: '$' + str(uuid.uuid1())[:8]
        handle = hid()
        if len(args):
            element_handle = hid()
            self._parent.execute_script(
                'window["{}"] = [];'.format(element_handle)
            )  # create element container in window scope
            for el in args:
                if isinstance(el, string_types):
                    # assume selector
                    self._parent.execute_script('window["{}"].push({});'.format(
                        element_handle, 'function() { return document.querySelector("%s") }' % el))
                else:
                    # assume web element
                    self._parent.execute_script(
                        'window["{}"].push(arguments[0]);'.format(element_handle), el)
            if len(args) == 1:
                condition = condition.replace('$el', 'window["{}"][0]{}'.format(
                    element_handle, '()' if isinstance(args[0], string_types) else ''))
            else:
                regex = r'(\$el\[([0-9]{0,3})\])'
                results = re.findall(regex, condition)  # [('$el[0]', '0'), ('$el[1]', '1'), ...]
                for result in results:
                    pos = eval(result[1])
                    if pos + 1 <= len(args):
                        condition = condition.replace(result[0], 'window["{}"][{}]{}'.format(
                            element_handle, pos, '()' if isinstance(args[pos], string_types) else ''))

            self._parent.execute_script(
                'window["%s"]=window.setInterval(function(){if(%s){ \
                (window.clearInterval(window["%s"])||true)&&(window["%s"]=-1); \
                delete window["%s"];}}, %s)' % (handle, condition, handle, handle, \
                                                element_handle, interval))  # create interval
        else:
            self._parent.execute_script(
                'window["%s"]=window.setInterval(function(){if(%s){ \
                (window.clearInterval(window["%s"])||true)&&(window["%s"]=-1);}}, %s)' % (
                    handle, condition, handle, handle, interval))  # create interval

        return handle

    def wait_status(self, handle):
        """
        :Description: Check the status of browser wait.
        :param handle: Interval handle returned from `self.wait`.
        :type handle: string
        :return: bool
        """
        return self._parent.execute_script('return window["%s"] == -1' % handle)

    def is_enabled(self):
        """
        Used to check at least one element is available and all are enabled.

        :return: bool
        """

        if self.get_property('disabled'):
            return False
        return True

    def is_disabled(self):
        """
        Used to check at least one element is available and all are disabled.

        :return: bool
        """
        if not self.get_property('disabled'):
            return False
        return True

    def is_visible(self):
        """
        :Description: Get's the visibility of the provided target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :return: bool
        """
        return bool(self._parent.execute_script(
            'return !!(arguments[0].offsetWidth || arguments[0].offsetHeight || \
            arguments[0].getBoundingClientRect().height || \
            arguments[0].getBoundingClientRect().width) && \
            (arguments[0].style.visibility == "" || arguments[0].style.visibility == "visible") && \
            (arguments[0].style.opacity ? arguments[0].style.opacity > 0 : true);',
            self))

    def click(self):
        """
        :Description: Execute the `click` event on the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        """
        actions = ActionChains(self._parent)
        actions.move_to_element(self).perform()
        super(Element, self).click()
        return self

    def submit(self):
        super(Element, self).submit()
        return self

    def clear(self):
        super(Element, self).clear()
        return self

    def send_keys(self, *value):
        super(Element, self).send_keys(*value)
        return self

    def dbl_click(self):
        """
        :Description: Execute the `dbclick` event on the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        """
        self._parent.execute_script(
            'var e = document.createEvent("mouseEvent"); \
            e.initEvent("dblclick", true, true); \
            arguments[0].dispatchEvent(e);', self)
        return self

    def select(self):
        """
        :Description: Sets the attribute `selected` to true and triggers `change` event.
        :param element: Element for browser instance to target.
        :type element: WebElement
        """
        self._parent.execute_script(
            'arguments[0].selected = "selected"; \
             arguments[0].dispatchEvent(new Event("change"));', self)

        return self

    def deselect(self):
        """
        :Description: Sets the attribute `selected` to null and triggers `change` event.
        :param element: Element for browser instance to target.
        :type element: WebElement
        """
        self._parent.execute_script(
            'arguments[0].selected = null; \
             arguments[0].dispatchEvent(new Event("change"));', self)

        return self

    def get_attribute(self, attribute, convert_type=True):
        """
        :Description: Return the given attribute of the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param attribute: Attribute of target element to return.
        :type attribute: string
        :param convert_type: If enabled, will return pythonic type.
        :type convert_type: bool
        :return: None, bool, int, float, string
        """
        attribute = self._parent.execute_script(
            'return arguments[0].getAttribute("%s");' % attribute, self)

        return self.__type2python(attribute) if convert_type else attribute

    def set_attribute(self, attribute, value):
        """
        :Description: Modify the given attribute of the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param attribute: Attribute of target element to modify.
        :type attribute: string
        :param value: Value of target element's attribute to modify.
        :type value: None, bool, int, float, string
        """
        self._parent.execute_script('arguments[0].setAttribute("%s", %s);' % (
            attribute, self.__type2js(value=value)), self)
        return self

    def remove_attribute(self, attribute):
        """
        :Description: Remove the given attribute from the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param attribute: Attribute of target element to remove.
        :type attribute: string
        """
        self._parent.execute_script('arguments[0].removeAttribute("%s");' % attribute, self)
        return self

    def get_property(self, prop):
        """
        :Description: Return the given attribute of the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param prop: Property of target element to return.
        :type prop: string
        :return: None, bool, int, float, string
        """
        return self._parent.execute_script('return arguments[0]["%s"];' % prop, self)

    def set_property(self, prop, value):
        """
        :Description: Modify the given attribute of the target element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param prop: Property of target element to modify.
        :type prop: string
        :param value: Value of target element's property to modify.
        :type value: None, bool, int float, string
        """
        self._parent.execute_script(
            'arguments[0]["%s"] = %s' % (prop, self.__type2js(value=value)), self)

        return self

    def get_value(self):
        """
        :Description: Return the value of the given element.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :return: string
        """
        return self.get_property(prop='value')

    def get_text(self):
        """
        :Description: Return the text content between the tags of the given element.
        :Info: Helpful for reading text from elements that do not support the `value` property.
        :param element: Element for browser instance to target.
        :return: string
        """
        return self.get_property(prop='innerText')

    def get_raw_text(self):
        """
        :Description: Return the text content between the tags of the given element.
        :Info: Helpful for reading text from elements that do not support the `value` property.
        :Warning: This will return the raw text content of the DOM's child scope.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :return: string
        """
        return self.get_property(prop='innerHTML')

    def trigger_event(self, event, event_type=None, options=None):
        """
        :Description: Trigger specified event of the given element.
        :param element: Element for browser instance to target.
        :type element: WebElement, (WebElement, ...)
        :param event: Event to trigger from target element.
        :type event: string, (string, ...)
        :param event_type: Event type.
        :type event_type: string
        :example: 'KeyboardEvent'
        :param options: Event options.
        :example: { 'bubbles': True, 'cancelable': False }
        :type options: dict
        """
        if not isinstance(self, (tuple, list)):
            element = [self]
        if not isinstance(event, (tuple, list)):
            event = [event]
        for el in element:
            for e in event:
                self._parent.execute_script(
                    'e = new %s("%s"); ops = %s; if (ops) {for(key in ops) { \
                        Object.defineProperty(e, key, { value: ops[key], configurable: true }) \
                    }} arguments[0].dispatchEvent(e)' % (
                        event_type if event_type else 'Event',
                        e, json.dumps(options) if options else 'undefined'
                    ), el)

    def scroll_into_view(self):
        """
        :Description: Scroll the target element into view.
        :Warning: This method relies on JQuery.
        :param element: Element for browser instance to target.
        :type element: WebElement
        """
        self._parent.execute_script('arguments[0].scrollIntoView(true);', self)
        return self

    @property
    def get_scrolling_offsets(self):
        """
        :Description: Returns the page scrolling x and y offsets.
        :return: dict
        """
        return {
            'x': self._parent.execute_script('return window.pageXOffset'),
            'y': self._parent.execute_script('return window.pageYOffset')
        }

    def ng_enable_debugging(self):
        """
        :Description: Enables angular debugging on given webpage.
        :Warning: This will only work for angular.js 1.x.
        """
        self._parent.execute_script('angular.reloadWithDebugInfo()')
        return self

    def ng_get_text(self):
        """
        :Description: Will return the DOM's value, if not found will default to `innerHTML`.
        :Warning: This will only work for angular.js 1.x.
        :Warning: This will only work for angular elements.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :return: string
        """
        return self._parent.execute_script('return angular.element(arguments[0]).text();', self)

    def ng_set_text(self, text):
        """
        :Description: Will set a DOM's value, if not found will default to `innerHTML`.
        :Warning: This will only work for angular.js 1.x.
        :Warning: This will only work for angular elements.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param text: Text used for related operation.
        :type text: string
        """
        self._parent.execute_script('angular.element(arguments[0]).text("%s");' % text, self)

    def ng_toggle_class(self, target):
        """
        :Description: Toggle DOM class.
        :Warning: This will only work for angular.js 1.x.
        :Warning: This will only work for angular elements.
        :param element: Element for browser instance to target.
        :param target: Class to toggle.
        """
        self._parent.execute_script(
            'angular.element(arguments[0]).toggleClass("%s");' % target, self)
        return self

    def ng_trigger_event_handler(self, event):
        """
        :Description: Trigger angular event handler of element.
        :Warning: This will only work for angular.js 1.x.
        :Warning: This will only work for angular elements.
        :param element: Element for browser instance to target.
        :param event: Event to trigger.
        """
        self._parent.execute_script(
            'angular.element(arguments[0]).triggerHandler("%s");' % event, self)
        return self

    @staticmethod
    def __d2b_notation(prop):
        """
        :Description: Transform javascript dot notation to bracket notation.
        :param prop: Property to transform.
        :type prop: string
        :example: 'messages.total' >> someObject['messages']['total']
        :return: string
        """
        results = re.compile('[[$a-zA-Z]{0,}.').findall(prop)
        for i in range(0, len(results)):
            results[i] = ("['%s']" % results[i]).replace('.', '')
        return ''.join(results)

    @classmethod
    def __serialize_params(cls, params):
        param_str = ''
        for param in params:
            param_str += '%s,' % cls.__type2js(value=param)
        if param_str.endswith(','):
            param_str = param_str.replace(param_str[-1], '')
        return param_str

    def ng_get_scope_property(self, element, prop):
        """
        :Description: Will return value of property of element's scope.
        :Warning: This will only work for angular.js 1.x.
        :Warning: Requires angular debugging to be enabled.
        :param element: Element for browser instance to target.
        :param prop: Property of element's angular scope to target.
        :type prop: string
        :example: 'messages.total'
        :return: string
        """
        return self._parent.execute_script(
            'return angular.element(arguments[0]).scope()%s;' % self.__d2b_notation(
                prop=prop
            ), element)

    def ng_set_scope_property(self, element, prop, value):
        """
        :Description: Will set value of property of angular element's scope.
        :Warning: This will only work for angular.js 1.x.
        :Warning: Requires angular debugging to be enabled.
        :param element: Element for browser instance to target.
        :param prop: Property of element's angular scope to target.
        :type prop: string
        :example: 'messages.total'
        :param value: Value to specify to angular element's scope's property.
        :type value: None, bool, int, float, string
        """
        self._parent.execute_script(
            'angular.element(arguments[0]).scope()%s = %s;' % (
                self.__d2b_notation(prop=prop), self.__type2js(value=value)
            ), element)
        return self

    def ng_call_scope_function(self, func, params='', return_out=False):
        """
        :Description: Will execute scope function with provided parameters.
        :Warning: This will only work for angular.js 1.x.
        :Warning: Requires angular debugging to be enabled.
        :param element: Element for browser instance to target.
        :param func: Function to execute from angular element scope.
        :type func: string
        :param params: String (naked) args, or list of parameters to pass to target function.
        :type params: string, tuple, list
        :param return_out: Return output of function call otherwise None
        :type return_out: bool
        """
        if isinstance(params, string_types):
            param_str = params
        elif isinstance(params, (tuple, list)):
            param_str = self.__serialize_params(params)
        else:
            raise ValueError('Invalid type specified for function parameters')
        exec_str = 'angular.element(arguments[0]).scope().%s(%s);' % (func, param_str)
        if return_out:
            return self.__type2python(
                self._parent.execute_script('return {}'.format(exec_str), self))
        else:
            self._parent.execute_script(exec_str, self)

    def ng_get_ctrl_property(self, prop):
        """
        :Description: Will return value of property of element's controller.
        :Warning: This will only work for angular.js 1.x.
        :Warning: Requires angular debugging to be enabled.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param prop: Property of element's angular controller to target.
        :type prop: string
        :example: 'messages.total'
        :return: string
        """
        return self._parent.execute_script(
            'return angular.element(arguments[0]).controller()%s;' % self.__d2b_notation(prop=prop),
            self)

    def ng_set_ctrl_property(self, prop, value):
        """
        :Description: Will set value of property of element's controller.
        :Warning: This will only work for angular.js 1.x.
        :Warning: Requires angular debugging to be enabled.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param prop: Property of element's angular scope to target.
        :type prop: string
        :example: 'messages.total'
        :param value: Value to specify to angular element's controller's property.
        :type value: None, bool, int, float, string
        """
        self._parent.execute_script(
            'angular.element(arguments[0]).controller()%s = %s;' % (
                self.__d2b_notation(prop=prop), self.__type2js(value=value)
            ), self)

    def ng_call_ctrl_function(self, func, params='', return_out=False):
        """
        :Description: Will execute controller function with provided parameters.
        :Warning: This will only work for angular.js 1.x.
        :Warning: Requires angular debugging to be enabled.
        :param element: Element for browser instance to target.
        :param func: Function to execute from angular element controller.
        :type func: string
        :param params: String (naked) args, or list of parameters to pass to target function.
        :type params: string, tuple, list
        :param return_out: Return output of function call otherwise None
        :type return_out: bool
        """
        if isinstance(params, string_types):
            param_str = params
        elif isinstance(params, (tuple, list)):
            param_str = self.__serialize_params(params)
        else:
            raise ValueError('Invalid type specified for function parameters')
        exec_str = 'angular.element(arguments[0]).controller().%s(%s);' % (func, param_str)
        if return_out:
            return self.__type2python(
                self._parent.execute_script('return {}'.format(exec_str), self))
        else:
            self._parent.execute_script(exec_str, self)

    def ng2_get_component_property(self, prop):
        """
        :Description: Will get value of property of element's component instance.
        :Warning: This will only work for Angular components.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param prop: Property of element's component to target.
        :type prop: string
        :example: 'messages.total'
        :return: string
        """
        return self._parent.execute_script(
            'return ng.probe(arguments[0]).componentInstance%s;' % self.__d2b_notation(prop=prop),
            self)

    def ng2_set_component_property(self, prop, value):
        """
        :Description: Will set value of property of element's component instance.
        :Warning: This will only work for Angular components.
        :param element: Element for browser instance to target.
        :type element: WebElement
        :param prop: Property of element's component to target.
        :type prop: string
        :example: 'messages.total'
        :param value: Value to specify to component's property.
        :type value: None, bool, int, float, string
        """
        self._parent.execute_script(
            'ng.probe(arguments[0]).componentInstance%s = %s;' % (
                self.__d2b_notation(prop=prop), self.__type2js(value=value)), self)

        return self

    def ng2_call_component_function(self, func, params='', return_out=False):
        """
        :Description: Will execute the component instance function with provided parameters.
        :Warning: This will only work for Angular components.
        :param element: Element for browser instance to target.
        :param func: Function to execute from component instance.
        :type func: string
        :param params: String (naked) args, or list of parameters to pass to target function.
        :type params: string, tuple, list
        :param return_out: Return output of function call otherwise None
        :type return_out: bool
        """
        if isinstance(params, string_types):
            param_str = params
        elif isinstance(params, (tuple, list)):
            param_str = self.__serialize_params(params)
        else:
            raise ValueError('Invalid type specified for function parameters')
        exec_str = 'ng.probe(arguments[0]).componentInstance.%s(%s);' % (func, param_str)
        if return_out:
            return self.__type2python(
                self._parent.execute_script('return {}'.format(exec_str), self))
        else:
            self._parent.execute_script(exec_str, self)

    # def fmt(self, **kwargs):
    #     """
    #     Used to format selectors.
    #
    #     :return: Elements
    #     """
    #     self._selector = self.selector
    #     self.selector = Template(self.selector).safe_substitute(**kwargs)
    #     return self

    def mousedown(self):
        """
        Dispatches a mousedown event on the given element.

        :return: Element, None
        """
        # pragma: no cover
        # ignoring from coverage, assume covered in scroll_to and trigger_event
        if self:
            self.scroll_into_view()
            self.trigger_event(self, 'mousedown', 'MouseEvent')
            return self
        return None

    def wait_for(self, timeout, available=True, error=None):
        """
        Wait for a given element to become available.

        :param timeout: Time in seconds to wait for element.
        :type timeout: int
        :available: Used to check whether element is available or not available.
        :type available: bool
        :param error: Error message, if passed will raise NoSuchElementException.
        :type error: string, bool
        :return: Element, None
        """
        check = Checks(self)
        if not self._parent.wait(timeout=timeout, condition=check.available \
                if available else check.not_available):

            if error:
                raise NoSuchElementException(error if isinstance(error, string_types) else \
                                                 'Element by selector "{}" was {}' \
                                             .format(self.selector, 'not found' if available else 'found'))
            return None

        return self

    def wait_village(self, timeout=30):
        while not self.is_visible() and timeout > 0:
            time.sleep(1)
            timeout -= 1
        return self

    def wait_visible(self, timeout=10, error=None):
        """
        Wait for given element to be visible.

        :param timeout: Time in seconds to wait for element.
        :type timeout: int
        :param error: Raise ElementNotVisibleException on failure.
        :type error: bool, string
        :return: Element, None
        """
        if not self._parent.wait(timeout=timeout, condition=self.is_visible):
            if error:
                raise ElementNotVisibleException(error if isinstance(error, string_types) \
                                                     else f'Element by selector "{self.selector}"\
                                                      not found or is not visible')
            return None

        return self

    def wait_invisible(self, timeout, error=None):
        """
        Wait for given element to be invisible.

        :param timeout: Time in seconds to wait for element.
        :type timeout: int
        :param error: Raise InvalidElementStateException on failure.
        :type error: bool, string
        :return: Element, None
        """

        if not self._parent.wait(timeout=timeout, condition=not self.is_visible):
            if error:
                raise InvalidElementStateException(error if isinstance(error, string_types) else \
                                                       'Element by selector "{}" not found or is visible'.format(
                                                           self.selector))
            return None

        return self

    def wait_enabled(self, timeout, error=None):
        """
        Wait for given element to be enabled.

        :param timeout: Time in seconds to wait for element.
        :type timeout: int
        :param error: Raise InvalidElementStateException on failure.
        :type error: bool, string
        :return: Element, None
        """
        check = Checks(self)

        if not self._parent.wait(timeout=timeout, condition=check.enabled):
            if error:
                raise InvalidElementStateException(error if isinstance(error, string_types) else \
                                                       'Element by selector "{}" not found or is disabled'.format(
                                                           self.selector))
            return None

        return self

    def wait_disabled(self, timeout, error=None):
        """
        Wait for given element to be disabled.

        :param timeout: Time in seconds to wait for element.
        :type timeout: int
        :param error: Raise InvalidElementStateException on failure.
        :type error: bool, string
        :return: Element, None
        """
        check = Checks(self)
        if not self._parent.wait(timeout=timeout, condition=check.disabled):
            if error:
                raise InvalidElementStateException(error if isinstance(error, string_types) else \
                                                       'Element by selector "{}" not found or is enabled'.format(
                                                           self.selector))
            return None

        return self

    def wait_js(self, condition, interval):
        """
        Wait for element by javascript condition.

        :param condition: Javascript condition to execute.
        :type condition: string
        :param interval: Interval to check condition by in ms.
        :type interval: int
        :return: Element, None
        """
        self.wait_handle = self.wait(condition, interval, self.selector)
        return self

    def switch_to(self):
        """
        Switch into an iframe element context.

        :return: Element, None
        """
        found = self
        if found:
            self._parent.switch_to.frame(found)
            return self
        return None

    def wait_presense_of_element(self, timeout=20):
        wait = WebDriverWait(self, timeout)
        element = wait.until(EC.presence_of_element_located(self))
        return element


class Checks(Resource):
    """
    Base resource for multiple element checks.

    :param elements: Elements instance to reference.
    :type element: Elements
    """

    def __init__(self, elements, value=''):

        self.elements = elements
        self.validate()
        self.value = value

        try:
            self.elements = iter(elements)
        except TypeError:
            self.elements = [self.elements]

    def visible(self):
        """
        Used to check at least one element is available and all are visible.

        :return: bool
        """
        found = self
        return found and self.is_visible(found)

    def invisible(self):
        """
        Used to check at least one element is available and all are invisible.

        :return: bool
        """
        found = self.elements if self.elements else None
        if found:
            for element in found:
                if element.is_visible():
                    return False
            return True

    def is_enabled(self):
        """
        Used to check at least one element is available and all are enabled.

        :return: bool
        """
        found = self.elements
        if not len(found):  # pylint: disable=len-as-condition
            return False
        for element in found:
            if self.elements.get_property(element, 'disabled'):
                return False
        return True

    def is_disabled(self):
        """
        Used to check at least one element is available and all are disabled.

        :return: bool
        """
        found = self.elements
        if not len(found):  # pylint: disable=len-as-condition
            return False
        for element in found:
            if not self.elements.get_property(element, 'disabled'):
                return False
        return True

    # meta = {'required_fields': [('elements', Element)]}



import json
import os
import time
from functools import wraps
from math import floor
from selenium.common.exceptions import NoSuchElementException
# from pyseleniumjs import E2EJS
from selenium.common.exceptions import JavascriptException
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from tractordriver.WebElement import Element


#
# class Element(WebElement):
#     def __init__(self, parent):
#         super(Element, self).__init__(parent, id_=None, w3c=False)
#
#     def __repr__(self):
#         return '<{0.__module__}.{0.__name__} (session="{1}", element="{2}")>'.format(
#             type(self), self._parent.session_id, self._id)


class AngularNotFoundException(WebDriverException):
    """An exception raised if angular.js could not be found on a web page."""


def angular_wait_required(wrapped):
    @wraps(wrapped)
    def wait_for_angular(driver, *args, **kwargs):
        driver.wait_for_angular()
        return wrapped(driver, *args, **kwargs)

    return wait_for_angular


class By(By):
    MODEL = "model name"
    BINDING = "Binding"
    BUTTON_TEXT = "Button text"
    PARTIAL_BUTTON_TEXT = "Partial button text"
    REPEATER_ROWS = 'repeater_rows'


class TractorDriver(object):

    def __init__(self, *args, **kwargs):
        super(TractorDriver, self).__init__()
        self._web_element_cls = Element
        file_source = os.path.join(os.path.dirname(__file__), 'clientsidescripts.js')
        with open(file_source) as f:
            js_script = f.read()
            f.close()
        # self._web_element_cls.js = E2EJS
        self._exports = json.loads(self.execute_script(js_script + """return JSON.stringify(exports);"""))
        self.__dict__.update(self._exports)

    # @staticmethod
    def explicity_wait_for_visible(self, element, timeout=60):
        wait = WebDriverWait(self, timeout=timeout)
        result = wait.until(EC.visibility_of_element_located(element))
        # element = wait.until(EC.presence_of_element_located(element))
        return result

    def wait_for_clickable(self, element, timeout=5):
        wait = WebDriverWait(self, timeout=timeout)
        element = wait.until(EC.element_to_be_clickable(element))
        return element

    def wait_for(self, wait_type, element, timeout=5):
        """
        :param wait_type: method
        :param timeout: int
        :param element: Element
        """
        wait = WebDriverWait(self, timeout)
        element = wait.until(wait_type(element))
        return element

    def find_element(self, by=By.ID, value=None, auto_wait=True, timeout=20):
        """
        :param auto_wait: bool
        :param timeout: int
        :param wait_method: object
        :type by: By
        :type value: str
        """
        if by is By.MODEL:
            element = self.find_element_by_model(value)
        elif by is By.BINDING:
            element = self.find_element_by_binding(value)
        elif by is By.BUTTON_TEXT:
            element = self.find_element_by_button_text(value)
        elif by is By.PARTIAL_BUTTON_TEXT:
            element = self.find_element_by_partial_button_text(value)
        elif by is By.REPEATER_ROWS:
            element = self.find_all_repeater_rows(value)
        else:
            element = super().find_element(by=by, value=value)
        if isinstance(element, (list,)):
            for elem in element:
                elem.selector = value
        else:
            element.selector = value
        if auto_wait:
            if not isinstance(element, (list,)):
                element.scroll_into_view()
                element.wait_visible(timeout=timeout, error=True)
        return element

    def find_elements(self, by=By.ID, value=None, auto_wait=True, timeout=20):
        """
        :param auto_wait: bool
        :param timeout: int
        :param wait_method: object
        :type by: By
        :type value: str
        """
        if by is By.MODEL:
            self.find_elements_by_model(value)
        elif by is By.BINDING:
            elements = self.find_elements_by_binding(value)
        elif by is By.BUTTON_TEXT:
            elements = self.find_elements_by_button_text(value)
        elif by is By.PARTIAL_BUTTON_TEXT:
            elements = self.find_elements_by_partial_button_text(value)
        elif by is By.REPEATER_ROWS:
            elements = self.find_all_repeater_rows(value)
        else:
            elements = super().find_elements(by=by, value=value)

        if isinstance(elements, (list,)):
            for elem in elements:
                elem.selector = value
        else:
            elements.selector = value
        if auto_wait:
            if not isinstance(elements, (list,)):
                # element.scroll_into_view()
                elements.wait_visible(timeout=timeout, error=True)
        return elements

    def _run_script(self, script, *args, **kwargs):
        _async = kwargs.pop('_async', False)
        new_args = tuple(['true' if x is True else x for x in list(args)]) if True in args else args
        if _async:
            return self.execute_async_script(script, *new_args, **kwargs), self._web_element_cls
        else:
            return self.execute_script(script, *new_args, **kwargs)

    def wait_for_angular(self):
        return self._run_script(self.waitForAngular, 'body', _async=True)

    @angular_wait_required
    def find_elements_by_model(self, model):
        return self._run_script(self.findByModel, model, _async=False)

    @angular_wait_required
    def find_element_by_model(self, model):
        return self.find_elements_by_model(model)[0]

    @angular_wait_required
    def find_elements_by_binding(self, bindings):
        return self._run_script(self.findBindings, bindings)

    @angular_wait_required
    def find_element_by_binding(self, binding):
        return self.find_elements_by_binding(binding)[0]

    @angular_wait_required
    def find_elements_by_option(self, options):
        return self._run_script(self.findByOptions, options)

    @angular_wait_required
    def find_element_by_option_value(self, option, value, show_first=True):
        opts = self.find_elements_by_option(option)
        results = []
        for item in opts:
            if value in item.get_text():
                results.append(item)
        if not len(results):
            raise NoSuchElementException(f"could not find element by option {option} and value {value}")
        return results[0] if show_first else results

    @angular_wait_required
    def find_element_by_option(self, option):
        elements = self.find_elements_by_option(option)
        if len(elements) == 1:
            return elements[0]
        else:
            raise NoSuchElementException(f" {len(elements)} elements found with \"{option}\" option value. ")

    @angular_wait_required
    def find_elements_by_button_text(self, button_text):
        return self._run_script(self.findByButtonText, button_text)

    @angular_wait_required
    def find_element_by_button_text(self, button_text):
        elements = self.find_elements_by_button_text(button_text)
        if len(elements) == 1:
            return elements[0]
        else:
            raise NoSuchElementException(f" {len(elements)} elements found with \"{button_text}\" button text. ")

    @angular_wait_required
    def find_elements_by_partial_button_text(self, partial_button_text):
        return self._run_script(self.findByPartialButtonText, partial_button_text)

    @angular_wait_required
    def find_element_by_partial_button_text(self, partial_button_text):
        return self.find_elements_by_partial_button_text(partial_button_text)[0]

    @angular_wait_required
    def find_element_by_repeater_rows(self, repeater_rows, using=None):
        return self._run_script(self.findRepeaterRows, repeater_rows, using)

    @angular_wait_required
    def find_all_repeater_rows(self, repeater_rows):
        return self._run_script(self.findAllRepeaterRows, repeater_rows)

    @angular_wait_required
    def find_all_repeater_rows_by_value(self, repeater_rows, value, exact_match=False):
        elements = []
        items = self.find_all_repeater_rows(repeater_rows)
        for item in items:
            if exact_match:
                if item.get_text() == value:
                    elements.append(item)
            else:
                if value in item.get_text():
                    elements.append(item)
        if not len(elements):
            raise NoSuchElementException(f'Could not found any element by {repeater_rows} repeater and {value} value')
        return elements

    @angular_wait_required
    def find_repeater_row_by_value(self, repeater_rows, value, exact_match=False):
        """
        :type repeater_rows: str
        :type value: str
        :type exact_match: bool
        """
        items = self.find_all_repeater_rows_by_value(repeater_rows, value, exact_match=exact_match)
        if len(items) != 1:
            raise NoSuchElementException(
                f"Expected to find one item with {repeater_rows} repeater and {value} value but multiple items have "
                f"been found")
        return items[0]

    @angular_wait_required
    def find_all_repeater_column(self, repeater_column):
        return self._run_script(self.findRepeaterColumn, repeater_column)

    def repeater_match(self):
        return self._run_script(self.repeaterMatch)

    @angular_wait_required
    def find_element_by_css_containing_text(self, css_containing_text):
        return self._run_script(self.findByCssContainingText, css_containing_text)

    def _test_for_angular(self, timeout=10):
        return self._run_script(self.testForAngular, floor(timeout / 1000))[0]

    def reload(self):
        url = self.execute_script('return window.location.href')
        self.get(url)

    @angular_wait_required
    def set_location(self, url):
        return self._run_script(self.setLocation, 'body', url)

    def _location_equals(self, location):
        result = self.execute_script('return window.location.href')
        return result == location

    # @property
    # @angular_wait_required
    # def current_url(self):
    #     return self.current_url

    @property
    @angular_wait_required
    def current_location(self):
        return self.execute_script('return window.location.href')

    @property
    @angular_wait_required
    def page_source(self):
        return self.page_source

    # TODO: fix issue catching exception
    def is_angular_page(self, timeout=10):

        # test_result = self._test_for_angular()
        # print(test_result)
        # angular_on_page = test_result[0]
        # if not angular_on_page:
        #     message = test_result[1]
        #     raise AngularNotFoundException(
        #         'Angular could not be found on page: {}:'
        #         ' {}'.format(self.current_url, message)
        #     )
        try:
            self._test_for_angular(timeout)
        except JavascriptException:
            return False
        return True

    def wait_and_click(self, element, timeout=10):
        wait = WebDriverWait(self, timeout)
        element = wait.until(EC.element_to_be_clickable(element))
        element.click()
        return element

    def exit(self, safe_exit=False):
        """
        Safely exit instance of webdriver.

        :param safe_exit: Disable any possible alert or confirmation popup windows.
        :type safe_exit: bool
        """
        if safe_exit:
            self.execute_script('delete window.alert; delete window.confirm')
        try:
            self.browser.stop_client()
        except (WebDriverException, AttributeError):
            self.logger.warning('Could not close remote driver')
        finally:
            self.quit()

    # TODO: Add browser logs to the Aluure report
    def browser_logs(self, name=None, path=None):
        """
        Dumps browser logs to local directory.

        :Warning: `self.js.console_logger` must be executed to store logs.
        :param name: Name log file dropped to disk, will default to timestamp if not specified.
        :type name: string
        :param path: Path to drop console log in.
        :type path: string
        :return: string
        """
        if path and not os.path.exists(path):
            os.mkdir(path)
        try:
            timestamp = str(int(time.time()))
            log_path = '%sconsole.%s.json' % (path, ('%s.%s' % (name, timestamp)) if \
                name else timestamp)
            with open(log_path, 'a', encoding='utf-8') as logfile:
                logfile.write(self.console_dump())
            return log_path
        except WebDriverException:
            self.logger.critical('Browser logger object not found, could not return any logs.')

    def screen_shot(self, prefix=None, path=None):
        """
        Takes a screen shot and saves it specified path.

        :param prefix: Prefix for screenshot.
        :type prefix: string
        :param path: Path to drop screen shot in.
        :type path: string
        :return: string
        """
        file_location = os.path.join(
            path if path else './', (prefix + '_' if prefix else '') + str(time.time()) + '.png')
        self.browser.get_screenshot_as_file(filename=file_location)
        return file_location

    @classmethod
    def wait(cls, timeout=1, condition=None, reverse=False, throw_error=False):
        """
        Assisted delays between browser and main thread.

        :param timeout: Time in seconds to wait.
        :type timeout: int
        :param condition: (callable) Wait 1 to timeout seconds until condition met.
        :param reverse: Will wait for the condition to evaluate to False instead of True.
        :param throw_error: Will throw error raised by condition at end of timeout.
        :type throw_error: bool
        :return: bool
        """
        if callable(condition):
            if not isinstance(timeout, int) or timeout < 1:
                raise ValueError('Timeout must be an integer or float greater than or equal to 1')
            error = None
            for _ in range(timeout):
                try:
                    if reverse:
                        if not condition():
                            return False
                    else:
                        if condition():
                            return True
                except Exception as exc:  # pylint: disable=broad-except
                    if throw_error:
                        error = exc
                time.sleep(1)
            if error and throw_error:
                raise error  # pylint: disable=raising-bad-type
            return reverse
        else:
            time.sleep(timeout)
            return True

    def console_logger(self):
        """
        :Description: Override browser console for log and error store.
        :Source: https://github.com/neetVeritas/pyselenium-js/issues/9#issuecomment-304284471
        :Warning: This will only enable console logging per session.
        """
        self.execute_script(
            'var _0x4f63=["\x24\x63\x6F\x6E\x73\x6F\x6C\x65","\x24\x6C\x6F\x67\x73","\x24\x65\x72\x72\x6F\x72\x73","\x24\x77\x61\x72\x6E\x69\x6E\x67\x73","\x61\x73\x73\x65\x72\x74","\x63\x6C\x65\x61\x72","\x67\x72\x6F\x75\x70","\x67\x72\x6F\x75\x70\x43\x6F\x6C\x6C\x61\x70\x73\x65\x64","\x67\x72\x6F\x75\x70\x45\x6E\x64","\x69\x6E\x66\x6F","\x74\x61\x62\x6C\x65","\x74\x69\x6D\x65","\x74\x69\x6D\x65\x45\x6E\x64","\x6C\x6F\x67","\x70\x72\x6F\x74\x6F\x74\x79\x70\x65","\x70\x75\x73\x68","\x65\x72\x72\x6F\x72","\x65\x78\x63\x65\x70\x74\x69\x6F\x6E","\x74\x72\x61\x63\x65","\x77\x61\x72\x6E","\x64\x75\x6D\x70","\x73\x74\x72\x69\x6E\x67\x69\x66\x79","\x24","\x61\x6A\x61\x78\x45\x72\x72\x6F\x72","\x75\x72\x6C","\x74\x79\x70\x65","\x73\x74\x61\x74\x75\x73","\x72\x65\x73\x70\x6F\x6E\x73\x65\x54\x65\x78\x74","\x63\x6F\x6E\x73\x6F\x6C\x65"];function Logger(){this[_0x4f63[0]]= console,this[_0x4f63[1]]= [],this[_0x4f63[2]]= [],this[_0x4f63[3]]= [],this[_0x4f63[4]]= this[_0x4f63[0]][_0x4f63[4]],this[_0x4f63[5]]= this[_0x4f63[0]][_0x4f63[5]],this[_0x4f63[6]]= this[_0x4f63[0]][_0x4f63[6]],this[_0x4f63[7]]= this[_0x4f63[0]][_0x4f63[7]],this[_0x4f63[8]]= this[_0x4f63[0]][_0x4f63[8]],this[_0x4f63[9]]= this[_0x4f63[0]][_0x4f63[9]],this[_0x4f63[10]]= this[_0x4f63[0]][_0x4f63[10]],this[_0x4f63[11]]= this[_0x4f63[0]][_0x4f63[11]],this[_0x4f63[12]]= this[_0x4f63[0]][_0x4f63[12]]}Logger[_0x4f63[14]][_0x4f63[13]]= function(){this[_0x4f63[1]][_0x4f63[15]](arguments[0]),this[_0x4f63[0]][_0x4f63[13]](arguments[0])},Logger[_0x4f63[14]][_0x4f63[16]]= function(){this[_0x4f63[2]][_0x4f63[15]](arguments[0]),this[_0x4f63[0]][_0x4f63[16]](arguments[0])},Logger[_0x4f63[14]][_0x4f63[17]]= function(){this[_0x4f63[2]][_0x4f63[15]](arguments[0]),this[_0x4f63[0]][_0x4f63[17]](arguments[0])},Logger[_0x4f63[14]][_0x4f63[18]]= function(){this[_0x4f63[2]][_0x4f63[15]](arguments[0]),this[_0x4f63[0]][_0x4f63[18]](arguments[0])},Logger[_0x4f63[14]][_0x4f63[19]]= function(){this[_0x4f63[3]][_0x4f63[15]](arguments[0]),this[_0x4f63[0]][_0x4f63[19]](arguments[0])},Logger[_0x4f63[14]][_0x4f63[20]]= function(){return JSON[_0x4f63[21]]({logs:this[_0x4f63[1]],errors:this[_0x4f63[2]],warnings:this[_0x4f63[3]]})},(window[_0x4f63[22]]&& $(document)&& $(document)[_0x4f63[23]]?!0:!1)&& $(document)[_0x4f63[23]](function(){console[_0x4f63[16]]({location:arguments[2][_0x4f63[24]],method:arguments[2][_0x4f63[25]],status:arguments[1][_0x4f63[26]],error:arguments[3],response:arguments[1][_0x4f63[27]]})}),window[_0x4f63[28]]=  new Logger')

    def console_dump(self):
        """
        :Description: Return console logs as stringified JSON structure.
        :Warning: This will only work once `console_logger` is executed.
        :return: string
        """
        return self.execute_script('return console.dump()')



from selenium import webdriver
from webdrivermanager import ChromeDriverManager, OperaChromiumDriverManager, GeckoDriverManager, EdgeDriverManager
# from webdriver_manager.firefox import GeckoDriverManager
# from webdriver_manager.microsoft import EdgeDriverManager
from webdriver_manager.microsoft import IEDriverManager
import os
from tractordriver.TractorScripts import TractorDriver


class Firefox(TractorDriver, webdriver.Firefox):
    def __init__(self, manual_mode=False, firefox_profile=None, firefox_binary=None,
                 timeout=30, capabilities=None, proxy=None,
                 executable_path="geckodriver", options=None,
                 log_path="geckodriver.log", firefox_options=None,
                 service_args=None, desired_capabilities=None):
        if not manual_mode:
            executable_path = GeckoDriverManager().download_and_install()[0]
            os.environ["webdriver.firefox.driver"] = executable_path
        super(Firefox, self).__init__(firefox_profile=firefox_profile, firefox_binary=firefox_binary,
                                      timeout=timeout, capabilities=capabilities, proxy=proxy,
                                      executable_path=executable_path, options=options,
                                      log_path=log_path, firefox_options=firefox_options,
                                      service_args=service_args, desired_capabilities=desired_capabilities)


class Chrome(TractorDriver, webdriver.Chrome):
    def __init__(self, executable_path=None, port=0,
                 options=None, service_args=None,
                 desired_capabilities=None, service_log_path=None,
                 chrome_options=None, version='latest'):
        if not executable_path:
            executable_path = ChromeDriverManager().download_and_install(version=version)[0]
            # os.environ["webdriver.chrome.driver"] = executable_path
        super(Chrome, self).__init__(executable_path=executable_path, port=port,
                                     options=options, service_args=service_args,
                                     desired_capabilities=desired_capabilities, service_log_path=service_log_path,
                                     chrome_options=chrome_options)


class Safari(TractorDriver, webdriver.Safari):
    pass


class Ie(TractorDriver, webdriver.Ie):
    def __init__(self, executable_path=None):
        if not executable_path:
            executable_path = IEDriverManager().install()
            os.environ["webdriver.ie.driver"] = executable_path
        super(Ie, self).__init__(executable_path=executable_path)


class Edge(TractorDriver, webdriver.Edge):

    def __init__(self, executable_path=None):
        if not executable_path:
            executable_path = EdgeDriverManager().download_and_install()[0]
            os.environ["webdriver.edge.driver"] = executable_path
        super(Edge, self).__init__(executable_path=executable_path)


class Opera(TractorDriver, webdriver.Opera):
    def __init__(self, executable_path=None,
                 desired_capabilities=None,
                 port=0,
                 service_log_path=None,
                 service_args=None,
                 options=None):
        if not executable_path:
            executable_path = OperaChromiumDriverManager().download_and_install()[0]
            os.environ["webdriver.opera.driver"] = executable_path
        super(Opera, self).__init__(executable_path=executable_path,
                                    desired_capabilities=desired_capabilities,
                                    port=port,
                                    service_log_path=service_log_path,
                                    service_args=service_args,
                                    options=options)


class BlackBerry(TractorDriver, webdriver.BlackBerry):
    pass


class PhantomJS(TractorDriver, webdriver.PhantomJS):
    pass


class Android(TractorDriver, webdriver.Android):
    pass


class WebKitGTK(TractorDriver, webdriver.WebKitGTK):
    pass


class Remote(TractorDriver, webdriver.Remote):
    pass
